/* mysqlConn.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

#include "mysqlConn.h"
#include "ConfigParser.h"
#include "tools.h"
#include "log.h"
#define DATABASE_VERSION "20"

static void show_mysql_error();

struct mysql_conn_struct {
    unsigned int tid;
    MYSQL *mysql_conn;
    struct used_result *used_results;
    struct escaped_string *escaped_strings;
    struct mysql_conn_struct *next;
};

struct used_result {
    MYSQL_RES *result;
    struct used_result *next;
};

struct escaped_string {
    char *string;
    struct escaped_string *next;
};

struct mysql_conn_struct *get_mysql_conn_struct();

struct mysql_conn_struct *mysql_conns = NULL;
static int mysql_serverport;
static char *mysql_host, *mysql_user, *mysql_pass, *mysql_base;

#ifdef HAVE_THREADS
static pthread_mutex_t synchronized;
#endif

static void check_mysql() {
    MYSQL *mysql_conn = get_mysql_conn();
    int errid;
    if((errid = mysql_ping(mysql_conn))) {
        if(mysql_errno(mysql_conn) == CR_SERVER_GONE_ERROR) {
            if(!mysql_real_connect(mysql_conn, mysql_host, mysql_user, mysql_pass, mysql_base, mysql_serverport, NULL, 0)) {
                show_mysql_error();
            }
        } else {
            //mysql error
            show_mysql_error();
        }
    }
}

MYSQL_RES *mysql_use() {
    struct mysql_conn_struct *mysql_conn = get_mysql_conn_struct();
    MYSQL_RES *res = mysql_store_result(mysql_conn->mysql_conn);
    struct used_result *result = malloc(sizeof(*result));
    if (!result) {
        mysql_free_result(res);
        return NULL;
    }
    result->result = res;
    result->next = mysql_conn->used_results;
    mysql_conn->used_results = result;
    return res;
}

void mysql_free() {
    struct mysql_conn_struct *mysql_conn = get_mysql_conn_struct();
    if(!mysql_conn) return;
    struct used_result *result, *next_result;
    for(result = mysql_conn->used_results; result; result = next_result) {
        next_result = result->next;
        mysql_free_result(result->result);
        free(result);
    }
    mysql_conn->used_results = NULL;
    struct escaped_string *escaped, *next_escaped;
    for(escaped = mysql_conn->escaped_strings; escaped; escaped = next_escaped) {
        next_escaped = escaped->next;
        free(escaped->string);
        free(escaped);
    }
    mysql_conn->escaped_strings = NULL;
}

int reload_mysql() {
    char *new_mysql_host = get_string_field("MySQL.host");
    char *new_mysql_user = get_string_field("MySQL.user");
    char *new_mysql_pass = get_string_field("MySQL.pass");
    char *new_mysql_base = get_string_field("MySQL.base");
    if(!(new_mysql_host && new_mysql_user && new_mysql_pass && new_mysql_base))
        return 0;
    
    //replace login data
    if(mysql_host)
        free(mysql_host);
    mysql_host = strdup(new_mysql_host);
    
    if(mysql_user)
        free(mysql_user);
    mysql_user = strdup(new_mysql_user);
    
    if(mysql_pass)
        free(mysql_pass);
    mysql_pass = strdup(new_mysql_pass);
    
    if(mysql_base)
        free(mysql_base);
    mysql_base = strdup(new_mysql_base);
    
    mysql_serverport = get_int_field("MySQL.port");
    if(!mysql_serverport)
        mysql_serverport = 3306;
    return 1;
}

void init_mysql() {
    THREAD_MUTEX_INIT(synchronized);
    
    MYSQL *mysql_conn = get_mysql_conn();
    
    //check database version...
    int version = 0;
    if(!mysql_query(mysql_conn, "SELECT `database_version` FROM `version`")) {
        MYSQL_RES *res = mysql_use();
        MYSQL_ROW row;
        if((row = mysql_fetch_row(res))) {
            version = atoi(row[0]);
        }
    }
    if(!version) {
        //CREATE DATABASE
        FILE *f = fopen("database.sql", "r");
        mysql_set_server_option(mysql_conn, MYSQL_OPTION_MULTI_STATEMENTS_ON);
        if (f) {
            char line[512];
            char query_buffer[8192];
            int query_buffer_pos = 0;
            while (fgets(line, sizeof(line), f)) {
                query_buffer_pos += sprintf(query_buffer + query_buffer_pos, " %s", line);
                if(line[(strlen(line) - 2)] == ';') {
                    if(mysql_query(mysql_conn, query_buffer))
                        show_mysql_error();
                    query_buffer_pos = 0;
                }
            }
            fclose(f);
        } else
            printf_log("main", LOG_ERROR | LOG_MYSQL, "File not found: database.sql");
        f = fopen("database.defaults.sql", "r");
        if (f) {
            char line[4096];
            char query_buffer[131072];
            int query_buffer_pos = 0;
            while (fgets(line, sizeof(line), f)) {
                query_buffer_pos += sprintf(query_buffer + query_buffer_pos, " %s", line);
                if(line[(strlen(line) - 2)] == ';') {
                    if(mysql_query(mysql_conn, query_buffer))
                        show_mysql_error();
                    query_buffer_pos = 0;
                }
            }
            fclose(f);
        } else
            printf_log("main", LOG_ERROR | LOG_MYSQL, "File not found: database.defaults.sql");
        do { 
            MYSQL_RES *res = mysql_store_result(mysql_conn); 
            mysql_free_result(res); 
        } while(!mysql_next_result(mysql_conn));
        mysql_set_server_option(mysql_conn, MYSQL_OPTION_MULTI_STATEMENTS_OFF);
        mysql_query(mysql_conn, "INSERT INTO `version` (`database_version`) VALUES ('" DATABASE_VERSION "')");
    }
    else if(version < atoi(DATABASE_VERSION)) {
        //UPDATE DATABASE
        FILE *f = fopen("database.upgrade.sql", "r");
        mysql_set_server_option(mysql_conn, MYSQL_OPTION_MULTI_STATEMENTS_ON);
        if (f) {
            char line[512];
            char query_buffer[8192];
            int query_buffer_pos = 0, use_querys = 0;
            sprintf(query_buffer, "-- version: %d", version);
            while (fgets(line, sizeof(line), f)) {
                if(use_querys) {
                    query_buffer_pos += sprintf(query_buffer + query_buffer_pos, " %s", line);
                    if(line[strlen(line) - 1] == ';') {
                        mysql_query(mysql_conn, query_buffer);
                        query_buffer_pos = 0;
                    }
                } else if(!stricmplen(query_buffer, line, strlen(query_buffer))) {
                    use_querys = 1;
                }
            }
            if(query_buffer_pos) {
                if(mysql_query(mysql_conn, query_buffer))
                    show_mysql_error();
            }
            fclose(f);
        } else
            printf_log("main", LOG_ERROR | LOG_MYSQL, "File not found: database.upgrade.sql");
        do { 
            MYSQL_RES *res = mysql_store_result(mysql_conn); 
            mysql_free_result(res); 
        } while(!mysql_next_result(mysql_conn));
        mysql_set_server_option(mysql_conn, MYSQL_OPTION_MULTI_STATEMENTS_OFF);
        mysql_query(mysql_conn, "UPDATE `version` SET `database_version` = '" DATABASE_VERSION "'");
    }
}

void free_mysql() {
    struct mysql_conn_struct *mysql_conn, *next;
    for(mysql_conn = mysql_conns; mysql_conn; mysql_conn = next) {
        next = mysql_conn->next;
        mysql_close(mysql_conn->mysql_conn);
        free(mysql_conn);
    }
    mysql_conns = NULL;
}

static void show_mysql_error() {
    MYSQL *mysql_conn = get_mysql_conn();
    //show mysql_error()
    printf_log("main", LOG_ERROR | LOG_MYSQL, "Error: %s\n", mysql_error(mysql_conn));
}

void printf_mysql_query(const char *text, ...) {
    MYSQL *mysql_conn = get_mysql_conn();
    va_list arg_list;
    char queryBuf[MYSQLMAXLEN];
    int pos;
    queryBuf[0] = '\0';
    va_start(arg_list, text);
    pos = vsnprintf(queryBuf, MYSQLMAXLEN - 2, text, arg_list);
    va_end(arg_list);
    if (pos < 0 || pos > (MYSQLMAXLEN - 2)) pos = MYSQLMAXLEN - 2;
    queryBuf[pos] = '\0';
    printf_log("main", LOG_MYSQL, "%s\n", queryBuf);
    if(mysql_query(mysql_conn, queryBuf)) {
        check_mysql();
        if(mysql_query(mysql_conn, queryBuf)) {
            show_mysql_error();
        }
    }
}

void printf_long_mysql_query(int len, const char *text, ...) {
    MYSQL *mysql_conn = get_mysql_conn();
    va_list arg_list;
    char queryBuf[len];
    int pos;
    queryBuf[0] = '\0';
    va_start(arg_list, text);
    pos = vsnprintf(queryBuf, len - 2, text, arg_list);
    va_end(arg_list);
    if (pos < 0 || pos > (len - 2)) pos = len - 2;
    queryBuf[pos] = '\0';
    printf_log("main", LOG_MYSQL, "%s\n", queryBuf);
    if(mysql_query(mysql_conn, queryBuf)) {
        check_mysql();
        if(mysql_query(mysql_conn, queryBuf)) {
            show_mysql_error();
        }
    }
}

char* escape_string(const char *str) {
    struct mysql_conn_struct *mysql_conn = get_mysql_conn_struct();
    struct escaped_string *escapedstr = malloc(sizeof(*escapedstr));
    if (!escapedstr) {
        return NULL;
    }
    char escaped[strlen(str)*2+1];
    mysql_real_escape_string(mysql_conn->mysql_conn, escaped, str, strlen(str));
    escapedstr->string = strdup(escaped);
    escapedstr->next = mysql_conn->escaped_strings;
    mysql_conn->escaped_strings = escapedstr;
    return escapedstr->string;
}

struct mysql_conn_struct *get_mysql_conn_struct() {
    SYNCHRONIZE(synchronized);
    struct mysql_conn_struct *mysql_conn;
    unsigned int tid;
    #ifdef HAVE_THREADS
    tid = (unsigned int) pthread_self_tid();
    #else
    tid = 1;
    #endif
    for(mysql_conn = mysql_conns; mysql_conn; mysql_conn = mysql_conn->next) {
        if(mysql_conn->tid == tid) {
            DESYNCHRONIZE(synchronized);
            return mysql_conn;
        }
    }
    mysql_conn = malloc(sizeof(*mysql_conn));
    mysql_conn->mysql_conn = mysql_init(NULL);
    mysql_conn->tid = tid;
    mysql_conn->used_results = NULL;
    mysql_conn->escaped_strings = NULL;
    mysql_conn->next = mysql_conns;
    mysql_conns = mysql_conn;
    if (!mysql_real_connect(mysql_conn->mysql_conn, mysql_host, mysql_user, mysql_pass, mysql_base, mysql_serverport, NULL, 0)) {
        //error
        show_mysql_error();
    }
    DESYNCHRONIZE(synchronized);
    return mysql_conn;
}

MYSQL *get_mysql_conn() {
    struct mysql_conn_struct *mysql_conn = get_mysql_conn_struct();
    return mysql_conn->mysql_conn;
}
